from .PinnacleErrors import PinnacleError, TableNotFoundError, RowNotFoundError, TableAlreadyExistsError, ColumnMismatchError, DeletingCommandError, AlreadyCalledError, NoCommandWasFoundError, ParsingError
from .PinnacleSerializer import *


__version__ = "Beta"
__author__ = "Chtaïba Omar"
__email__ = "omarchtaiba4@gmail.com"
__github__ = "https://github.com/Omar1CHTAIBA"
__X__ = "https://x.com/OmarChtaib90132"
__date__ = '2/9/2025'


class PinnacleDB:
    def __init__(self, db_name):
        self.db_name = db_name
        self.db_file_path = db_name + '.pdb'
        self.transaction_active = False
        self.querying = []
        if not os.path.exists(self.db_file_path):
            try:
                PicleSerializer.serialize({'tables': {}}, self.db_file_path)
            except Exception as e:
                PinnacleError(f'An error occurred while creating the database file:\n\n{e}!')

        try:
            self.db = PicleSerializer.deserialize(self.db_file_path)
        except Exception as e:
            PinnacleError(f'An error occurred while loading the database file:\n\n{e}!')

    def __str__(self):
        return f"Version: {__version__}\n" \
           f"Author: {__author__}\n" \
           f"Email: {__email__}\n" \
           f"GitHub: {__github__}\n" \
           f"X: {__X__}\n" \
           f"Date: {__date__}"

    def _sync(self):
        try:
            PicleSerializer.serialize(self.db, self.db_file_path)
        except Exception as e:
            PinnacleError(f'An error occurred while saving the database file:\n\n{e}!')

    def call(self):
        if self.transaction_active:
            raise AlreadyCalledError("Transaction already activated!")
        self.transaction_active = True

    def show_commands(self):
        if self.querying:
            for index, command in enumerate(self.querying):
                print(f"Index: {index}, Command: {command}")
        else:
            raise NoCommandWasFoundError("No command was found by show_commands method!")

    def delete_command(self, index):
        try:
            del self.querying[index]
        except Exception as e:
            DeletingCommandError(f"An error occurred while deleting the command {self.querying[index]} at index {index}:\n\n{e}!")

    def fold(self):
        if not self.querying:
            raise NoCommandWasFoundError("No commands to process in the transaction log!")

        try:
            self.transaction_active = False

            for i in self.querying:
                query = i.split()

                if i.startswith("CREATE"):
                    table = query[2].strip()
                    columns_str = ' '.join(query[5:]).strip()

                    try:
                        columns = literal_eval(columns_str)
                        self.create_table(table, columns)
                    except Exception as e:
                        raise ParsingError(f"Error parsing columns {columns_str} for table '{table}': {e}")

                elif i.startswith("SET"):
                    table = query[2].strip()
                    key = query[5].strip()
                    values_str = ' '.join(query[8:]).strip()

                    try:
                        values = literal_eval(values_str)
                        self.set(table, key, *values)
                    except Exception as e:
                        raise ParsingError(f"Error parsing values {values_str} for table '{table}', key '{key}': {e}")

                elif i.startswith("DELETE"):
                    table = query[2].strip()
                    key = query[-1].strip()
                    self.delete(table, key)

                elif i.startswith("CLEAR"):
                    self.clear()

                elif i.startswith("DELETE__"):
                    table = query[2]
                    self.delete_table(table)

                elif i.startswith("BACKUP"):
                    backup_name = query[-1]
                    self.backup(backup_name)

                elif i.startswith("RESTORE"):
                    backup_name = query[-1]
                    self.restore(backup_name)

            self.querying = []
        except KeyboardInterrupt:
            raise KeyboardInterrupt("The Method got stopped processing by a Key Interruption!")
        except Exception as e:
            raise PinnacleError(e)

    def create_table(self, table_name, columns):
        if self.transaction_active:
            self.querying.append(f"CREATE TABLE: {table_name} , Columns: {columns}")
        else:
            tables = self.db['tables']
            if table_name in tables:
                raise TableAlreadyExistsError(f"Table '{table_name}' already exists!")

            # Create a new table
            tables[table_name] = {
                'columns': columns,
                'rows': {},
            }
            self.db['tables'] = tables
            self._sync()

    def set(self, table_name, key, *values):
        if self.transaction_active:
            self.querying.append(f"SET Table: {table_name} , Key: {key} , Values: {values}")
        else:
            tables = self.db['tables']
            if table_name not in tables:
                tables[table_name] = {'columns': [], 'rows': {}}
            table = tables[table_name]
            if len(values) != len(table['columns']):
                raise ColumnMismatchError(f"Expected {len(table['columns'])} values, got {len(values)}")

            table['rows'][key] = values
            self.db['tables'] = tables
            self._sync()

    def delete(self, table_name, key):
        if self.transaction_active:
            self.querying.append(f"DELETE: Table: {table_name} , Key: {key}")
        else:
            tables = self.db['tables']
            if table_name not in tables or key not in tables[table_name]['rows']:
                raise RowNotFoundError(f"Row with key '{key}' not found in table '{table_name}'!")
            del tables[table_name]['rows'][key]
            self.db['tables'] = tables
            self._sync()

    def fetch(self, table_name, key):
        tables = self.db['tables']
        if table_name not in tables or key not in tables[table_name]['rows']:
            raise RowNotFoundError(f"Row with key '{key}' not found in table '{table_name}'!")
        return tables[table_name]['rows'][key]

    def len_tables(self):
        return len(self.db['tables'])

    def fetchtables(self):
        tables = self.db['tables']
        if len(tables) == 0:
            return "No tables was found in database!"
        table_names = list(tables.keys())
        return table_names


    def count(self, table_name):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")
        return len(tables[table_name]['rows'])

    def fetchall(self, table_name, limit=None, offset=None):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")
        rows = list(tables[table_name]['rows'].items())
        if offset:
            rows = rows[offset:]
        if limit:
            rows = rows[:limit]
        return rows

    def fetchrows(self, table_name):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")
        return tables[table_name]['rows']

    def fetchcolumns(self, table_name):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")
        return tables[table_name]['columns']

    def fetchkeys(self, table_name):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")

        return tuple(tables[table_name]['rows'].keys())

    def fetchvalues(self, table_name):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")

        return tuple(tables[table_name]['rows'].values())

    def get_column_names(self, table_name):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")
        return tables[table_name]['columns']

    def fetchrows_by_value(self, table_name, value):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")

        result = []
        for row_key, row_values in tables[table_name]['rows'].items():
            if value in row_values:
                result.append((row_key, row_values))
        return result

    def fetchrows_by_key(self, key):
        found = []
        for table_name, table in self.db['tables'].items():
            if key in table['rows']:
                found.append((table_name, table['rows'][key]))

        if not found:
            raise RowNotFoundError(f"Row with key '{key}' not found in any table!")
        return found

    def clear(self):
        if self.transaction_active:
            self.querying.append("CLEAR DATABASE")
        else:
            self.db['tables'] = {}
            self._sync()

    def delete_table(self, table_name):
        if self.transaction_active:
            self.querying.append(f"DELETE TABLE: {table_name}")
        else:
            tables = self.db['tables']
            if table_name not in tables:
                raise TableNotFoundError(f"Table '{table_name}' not found!")
            del tables[table_name]
            self.db['tables'] = tables
            self._sync()

    def print_table(self, columns, rows):
        header = " |||| " + " | ".join(columns) + " | "
        print(header)
        for row_key, row_values in rows.items():
            row_str = f"| {row_key} | " + " | ".join(map(str, row_values)) + " |"
            print(row_str)

    def export_to_csv(self, table_name, file_path):
        tables = self.db['tables']
        if table_name not in tables:
            raise TableNotFoundError(f"Table '{table_name}' not found!")
        try:
            import csv
        except Exception as e:
            PinnacleError(f"An error occurred while importing csv module {e}")

        with open(file_path, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            columns = ['ID'] + tables[table_name]['columns']
            writer.writerow(columns)
            for row_key, row_values in tables[table_name]['rows'].items():
                writer.writerow([row_key] + list(row_values))

    def backup(self, backup_name):
        if self.transaction_active:
            self.querying.append(f"BACKUP DATABASE TO {backup_name}")
        else:
            with open(backup_name + '.pkl', 'wb') as backup_file:
                PicleSerializer.serialize(self.db, backup_file)

    def restore(self, backup_name):
        if self.transaction_active:
            self.querying.append(f"RESTORE DATABASE FROM {backup_name}")
        else:
            with open(backup_name + '.pkl', 'rb') as backup_file:
                self.db = PicleSerializer.deserialize(backup_file)
            self._sync()

    def close(self):
        self._sync()

